@@ -1,9 +1,9 @@
 /*****
 *
-* Copyright (C) 2001, 2002 Yoann Vandoorselaere <yoann@prelude-ids.org>
-* All Rights Reserved
+* Copyright (C) 2001, 2002, 2003, 2004, 2005 PreludeIDS Technologies. All Rights Reserved.
+* Author: Yoann Vandoorselaere <yoann.v@prelude-ids.com>
 *
-* This file is part of the Prelude program.
+* This file is part of the Prelude-Manager program.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
@@ -24,8 +24,22 @@
 #ifndef _MANAGER_IDMEF_MESSAGE_SCHEDULER_H
 #define _MANAGER_IDMEF_MESSAGE_SCHEDULER_H
 
+typedef struct idmef_queue idmef_queue_t;
+
 int idmef_message_scheduler_init(void);
 void idmef_message_scheduler_exit(void);
-void idmef_message_schedule(prelude_msg_t *msg);
+
+int idmef_message_schedule(idmef_queue_t *queue, prelude_msg_t *msg);
+
+void idmef_message_process(idmef_message_t *idmef);
+
+idmef_queue_t *idmef_message_scheduler_queue_new(void);
+
+void idmef_message_scheduler_queue_destroy(idmef_queue_t *queue);
+
+
+void idmef_message_scheduler_stop_processing(void);
+
+void idmef_message_scheduler_start_processing(void);
 
 #endif /* _MANAGER_IDMEF_MESSAGE_SCHEDULER_H */
