@@ -22,6 +22,7 @@
 *****/
 
 #include "config.h"
+#include "libmissing.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -253,91 +254,112 @@ static int get_params(gnutls_session session, gnutls_params_type type, gnutls_pa
         }
         
         pthread_mutex_lock(&dh_regen_mutex);
-        
         ret = gnutls_dh_params_cpy(cpy, cur_dh_params);
+        pthread_mutex_unlock(&dh_regen_mutex);
+        
         if ( ret < 0 ) {
                 prelude_log(PRELUDE_LOG_WARN, "could not copy dh params for sessions: %s.\n", gnutls_strerror(ret));
                 gnutls_dh_params_deinit(cpy);
                 return -1;
         }
-
-        pthread_mutex_unlock(&dh_regen_mutex);
         
         st->deinit = 1;
         st->type = type;
         st->params.dh = cpy;
-        
+
         return 0;
 }
 
 
 
-static int handle_gnutls_error(gnutls_session session, server_generic_client_t *client, int ret)
+static int handle_gnutls_error(prelude_io_t *pio, gnutls_session session, server_generic_client_t *client, int ret,
+                               gnutls_alert_description *alert_desc)
 {
-        int last_alert;
+        int level;
+        const char *alert;
         
-        if ( ret == GNUTLS_E_AGAIN ) {
-                
-                ret = gnutls_record_get_direction(session);
-                if ( ret == 1 ) 
+        if ( ret == GNUTLS_E_AGAIN ) {                
+                if ( gnutls_record_get_direction(session) == 1 )
                         server_logic_notify_write_enable((server_logic_client_t *) client);
                 
                 return 0;
         }
+
+        else if ( ret == GNUTLS_E_INTERRUPTED )
+                return 1;
         
-        if ( ret == GNUTLS_E_WARNING_ALERT_RECEIVED || ret == GNUTLS_E_FATAL_ALERT_RECEIVED ) {
-                last_alert = gnutls_alert_get(session);
-                server_generic_log_client(client, PRELUDE_LOG_WARN, "TLS alert: %s.\n", gnutls_alert_get_name(last_alert));
+        else if ( ret == GNUTLS_E_WARNING_ALERT_RECEIVED ) {
+                alert = gnutls_alert_get_name(gnutls_alert_get(session));
+                server_generic_log_client(client, PRELUDE_LOG_WARN, "TLS alert from client: %s.\n", alert);
+                return 1;
+        }
+
+        else if ( ret == GNUTLS_E_FATAL_ALERT_RECEIVED ) {
+                alert = gnutls_alert_get_name(gnutls_alert_get(session));
+                server_generic_log_client(client, PRELUDE_LOG_WARN, "TLS fatal alert from client: %s.\n", alert);
+        }
+
+        else {
+                server_generic_log_client(client, PRELUDE_LOG_WARN, "TLS error: %s.\n", gnutls_strerror(ret));
+                if ( alert_desc && (ret = gnutls_error_to_alert(ret, &level)) > 0 )
+                        *alert_desc = (gnutls_alert_description) ret;
         }
-        
-        server_generic_log_client(client, PRELUDE_LOG_WARN, "TLS error: %s.\n", gnutls_strerror(ret));
         
         return -1;
 }
 
 
 
-static int verify_certificate(server_generic_client_t *client, gnutls_session session)
+static int verify_certificate(server_generic_client_t *client, gnutls_session session, gnutls_alert_description *alert)
 {
 	int ret;
         time_t now;
+        unsigned int status;
         prelude_log_t pri = PRELUDE_LOG_WARN;
         
-	ret = gnutls_certificate_verify_peers(session);
+	ret = gnutls_certificate_verify_peers2(session, &status);
 	if ( ret < 0 ) {
-                server_generic_log_client(client, pri, "TLS certificate error: %s.\n", gnutls_strerror(ret));
-                return ret;
-        }
-
-	if ( ret == GNUTLS_E_NO_CERTIFICATE_FOUND ) {
-		server_generic_log_client(client, pri, "TLS authentication error: client did not send any certificate.\n");
-                gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_CERTIFICATE_UNOBTAINABLE);
-                return ret;
-	}
-
-        if ( ret & GNUTLS_CERT_SIGNER_NOT_FOUND) {
-		server_generic_log_client(client, pri, "TLS authentication error: client certificate issuer is unknown.\n");
-                gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_UNKNOWN_CA);
+                server_generic_log_client(client, pri, "error verifying certificate: %s.\n", gnutls_strerror(ret));
                 return ret;
         }
         
-        if ( ret & GNUTLS_CERT_INVALID ) {
-                server_generic_log_client(client, pri, "TLS authentication error: client certificate is NOT trusted.\n");
-                gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_CERTIFICATE_UNKNOWN);
+        if ( status & GNUTLS_CERT_SIGNER_NOT_FOUND) {
+                *alert = GNUTLS_A_UNKNOWN_CA;
+                server_generic_log_client(client, pri, "TLS authentication error: client certificate issuer is unknown.\n");
                 return -1;
         }
+        
+        else if ( status & GNUTLS_CERT_REVOKED ) {
+                *alert = GNUTLS_A_CERTIFICATE_REVOKED;
+                server_generic_log_client(client, pri, "TLS authentication error: client certificate is revoked.\n");
+                return -1;
+        }
+        
+        else if ( status & GNUTLS_CERT_INVALID ) {
+                *alert = GNUTLS_A_CERTIFICATE_UNKNOWN;
+                server_generic_log_client(client, pri, "TLS authentication error: client certificate is NOT trusted.\n");
+                return -1;
+        }
+
+#ifdef GNUTLS_CERT_INSECURE_ALGORITHM
+        else if ( status & GNUTLS_CERT_INSECURE_ALGORITHM ) {
+                *alert = GNUTLS_A_INSUFFICIENT_SECURITY;
+                server_generic_log_client(client, pri, "TLS authentication error: client use insecure algorithm");
+                return -1;
+        }
+#endif
 
         now = time(NULL);
         
         if ( gnutls_certificate_activation_time_peers(session) > now ) {
+                *alert = GNUTLS_A_BAD_CERTIFICATE;
                 server_generic_log_client(client, pri, "TLS authentication error: client certificate not yet activated.\n");
-                gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_BAD_CERTIFICATE);
                 return -1;
         }        
 
         if ( gnutls_certificate_expiration_time_peers(session) < now ) {
+                *alert = GNUTLS_A_CERTIFICATE_EXPIRED;
                 server_generic_log_client(client, pri, "TLS authentication error: client certificate expired.\n");
-                gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_CERTIFICATE_EXPIRED);
                 return -1;
         }
                 
@@ -412,7 +434,7 @@ static int certificate_get_peer_analyzerid(server_generic_client_t *client, gnut
 
 
 
-int manager_auth_client(server_generic_client_t *client, prelude_io_t *pio)
+int manager_auth_client(server_generic_client_t *client, prelude_io_t *pio, gnutls_alert_description *alert)
 {
         int ret;
         uint64_t analyzerid;
@@ -443,20 +465,25 @@ int manager_auth_client(server_generic_client_t *client, prelude_io_t *pio)
         }
         
         do {
-                ret = gnutls_handshake(session);
-        } while ( ret < 0 && ret == GNUTLS_E_INTERRUPTED );
-
-        if ( ret < 0 )
-                return handle_gnutls_error(session, client, ret);
+                ret = gnutls_handshake(session);                
+                if ( ret == 0 )
+                        ret = 1;
+                
+        } while ( ret < 0 && (ret = handle_gnutls_error(pio, session, client, ret, alert)) == 1 );
         
-        ret = verify_certificate(client, session);
+        if ( ret <= 0 )
+                return ret;
+        
+        ret = verify_certificate(client, session, alert);
         if ( ret < 0 ) 
                 return -1;
 
         ret = certificate_get_peer_analyzerid(client, session, &analyzerid, &permission);
-        if ( ret < 0 )
+        if ( ret < 0 ) {
+                *alert = GNUTLS_A_BAD_CERTIFICATE;
                 return -1;
-
+        }
+        
         ret = server_generic_client_set_permission(client, permission);
         if ( ret < 0 )
                 return -1;
@@ -477,18 +504,21 @@ int manager_auth_disable_encryption(server_generic_client_t *client, prelude_io_
         gnutls_session session;
         
         session = prelude_io_get_fdptr(pio);
-        
+
         do {
                 ret = gnutls_bye(session, GNUTLS_SHUT_RDWR);
-        } while ( ret < 0 && ret == GNUTLS_E_INTERRUPTED );
-        
-        if ( ret < 0 )          
-                return handle_gnutls_error(session, client, ret);
+                if ( ret == 0 )
+                        ret = 1;
+                
+        } while ( ret < 0 && (ret = handle_gnutls_error(pio, session, client, ret, NULL)) == 1 );
 
+        if ( ret <= 0 )
+                return ret;
+        
         gnutls_deinit(session);
         prelude_io_set_sys_io(pio, prelude_io_get_fd(pio));
-
-        return 1;
+        
+        return ret;
 }
 
 
@@ -496,7 +526,7 @@ int manager_auth_disable_encryption(server_generic_client_t *client, prelude_io_
 int manager_auth_init(prelude_client_t *client, int dh_bits, int dh_lifetime)
 {
         int ret;
-        char keyfile[256], certfile[256];
+        char keyfile[PATH_MAX], certfile[PATH_MAX], crlfile[PATH_MAX];
         prelude_client_profile_t *cp = prelude_client_get_profile(client);
 
         if ( ! dh_bits )
@@ -539,6 +569,15 @@ int manager_auth_init(prelude_client_t *client, int dh_bits, int dh_lifetime)
 		return -1;
 	}
 
+        prelude_client_profile_get_tls_server_crl_filename(cp, crlfile, sizeof(crlfile));
+        if ( access(crlfile, R_OK) == 0 ) {
+                ret = gnutls_certificate_set_x509_crl_file(cred, crlfile, GNUTLS_X509_FMT_PEM);
+                if ( ret < 0 ) {
+                        prelude_log(PRELUDE_LOG_ERR, "%s: %s\n", crlfile, gnutls_strerror(ret));
+                        return -1;
+                }
+        }
+        
         gnutls_dh_params_init(&cur_dh_params);
 	
 	ret = access(MANAGER_RUN_DIR, R_OK|W_OK);
