@@ -57,11 +57,9 @@
 extern manager_config_t config;
 
 prelude_client_t *manager_client;
-server_generic_t *sensor_server = NULL;
 
-static size_t nserver = 0;
 static char **global_argv;
-static volatile sig_atomic_t got_sighup = 0;
+static volatile sig_atomic_t got_signal = 0;
 
 
 
@@ -70,58 +68,19 @@ static volatile sig_atomic_t got_sighup = 0;
  */
 static void handle_signal(int sig) 
 {
-        /*
-         * re-establish signal handler.
-         */
-        signal(sig, handle_signal);
+        size_t i;
         
-        prelude_log(PRELUDE_LOG_INFO, "Caught signal %d.\n", sig);
-
         /*
          * stop the sensor server.
          */
-        sensor_server_stop(sensor_server);
+        for ( i = 0; i < config.nserver; i++ )
+                sensor_server_stop(config.server[i]);
+
+        got_signal = sig;
 }
 
 
 
-static void handle_sighup(int signo)
-{
-        /*
-         * re-establish signal handler.
-         */
-        signal(signo, handle_sighup);
-
-        /*
-         * stop the sensor server.
-         */
-        sensor_server_stop(sensor_server);
-        
-        got_sighup = 1;
-}
-
-
-
-static void init_manager_server(void) 
-{
-        int ret;
-        
-        nserver++;
-        
-        ret = manager_auth_init(manager_client, config.dh_bits, config.dh_regenerate);
-	if ( ret < 0 )
-                exit(1);
-        
-        ret = server_generic_bind((server_generic_t *) sensor_server, config.addr, config.port);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_WARN, "Error initializing prelude-manager server.\n");
-                exit(1);
-        }
-}
-
-
-
-
 static void restart_manager(void) 
 {
         int ret;
@@ -179,7 +138,6 @@ static void heartbeat_cb(prelude_client_t *client, idmef_message_t *idmef)
 int main(int argc, char **argv)
 {
         int ret;
-        prelude_string_t *err;
         struct sigaction action;
         prelude_option_t *manager_root_optlist;
         
@@ -190,8 +148,12 @@ int main(int argc, char **argv)
         
         /*
          * make sure we ignore sighup until acceptable.
-         */
+         */        
+#ifdef SA_INTERRUPT
+        action.sa_flags = SA_INTERRUPT;
+#else
         action.sa_flags = 0;
+#endif
         action.sa_handler = SIG_IGN;
         sigemptyset(&action.sa_mask);
         sigaction(SIGHUP, &action, NULL);
@@ -222,7 +184,6 @@ int main(int argc, char **argv)
         }
         prelude_log(PRELUDE_LOG_DEBUG, "- Initialized %d filtering plugins.\n", ret);
         
-        sensor_server = sensor_server_new();
         
         ret = manager_options_init(manager_root_optlist, &argc, argv);
         if ( ret < 0 )
@@ -241,25 +202,14 @@ int main(int argc, char **argv)
         
         ret = prelude_client_init(manager_client);
         if ( ret < 0 ) {
-                prelude_perror(ret, "error starting prelude-client");
-                
-                if ( prelude_client_is_setup_needed(ret) )
-                        prelude_client_print_setup_error(manager_client);
-
+                prelude_perror(ret, "error initializing prelude-client");
                 return ret;
         }
-        
-        ret = prelude_option_read(manager_root_optlist, &config.config_file, &argc, argv, &err, manager_client);        
-        if ( ret < 0 ) {
-                if ( err )
-                        prelude_log(PRELUDE_LOG_WARN, "Option error: %s.\n", prelude_string_get_string(err));
 
-                else if ( prelude_error_get_code(ret) != PRELUDE_ERROR_EOF )
-                        prelude_perror(ret, "error processing prelude-manager options");
-                
+        ret = manager_options_read(manager_root_optlist, &argc, argv);
+        if ( ret < 0 )
                 return -1;
-        }
-                
+        
         ret = prelude_client_start(manager_client);
         if ( ret < 0 ) {
                 prelude_perror(ret, "error starting prelude-client");                
@@ -284,35 +234,34 @@ int main(int argc, char **argv)
         /*
          * start server
          */
-        init_manager_server();
+        ret = manager_auth_init(manager_client, config.dh_bits, config.dh_regenerate);
+	if ( ret < 0 )
+                return -1;
         
         /*
          * setup signal handling
          */
-        action.sa_flags = 0;
-        sigemptyset(&action.sa_mask);
-        action.sa_handler = handle_signal;
+        sigaction(SIGPIPE, &action, NULL);
         
-        signal(SIGPIPE, SIG_IGN);
+        action.sa_handler = handle_signal;
         sigaction(SIGINT, &action, NULL);
         sigaction(SIGTERM, &action, NULL);
         sigaction(SIGABRT, &action, NULL);
         sigaction(SIGQUIT, &action, NULL);
-        
-        action.sa_handler = handle_sighup;
         sigaction(SIGHUP, &action, NULL);
         
-        server_generic_start(&sensor_server, nserver);
-        
+        server_generic_start(config.server, config.nserver);
                 
         /*
-         * we won't get here unless a signal is caught.
+         * we won't get there unless a signal is caught.
          */
-        server_generic_close(sensor_server);
+        if ( got_signal )
+                prelude_log(PRELUDE_LOG_WARN, "signal %d received, %s prelude-manager.\n",
+                            got_signal, (got_signal == SIGHUP) ? "will restart" : "terminating");
         
         idmef_message_scheduler_exit();
 
-        if ( got_sighup )
+        if ( got_signal == SIGHUP )
                 restart_manager();
         
         if ( config.pidfile )
