@@ -91,7 +91,7 @@ static int send_auth_result(server_generic_client_t *client, int result)
         int ret;
         uint64_t nident;
         prelude_client_profile_t *cp;
-        
+
         if ( ! client->msg ) {
                 ret = prelude_msg_new(&client->msg, 1, sizeof(uint64_t), PRELUDE_MSG_AUTH, 0);
                 if ( ret < 0 )
@@ -123,6 +123,27 @@ static int send_auth_result(server_generic_client_t *client, int result)
 
 
 
+static int send_queued_alert(server_generic_client_t *client)
+{
+        int ret;
+
+        do {
+                ret = gnutls_alert_send(prelude_io_get_fdptr(client->fd), GNUTLS_AL_FATAL, client->alert);
+        } while ( ret < 0 && ret == GNUTLS_E_INTERRUPTED );
+                
+        if ( ret == GNUTLS_E_AGAIN ) {
+                server_logic_notify_write_enable((server_logic_client_t *) client);
+                return 0;
+        }
+
+        client->alert = 0;
+        gnutls_deinit(prelude_io_get_fdptr(client->fd));
+        prelude_io_set_sys_io(client->fd, prelude_io_get_fd(client->fd));
+
+        return 1;
+}
+
+
 
 /*
  * Read the message sent by the Prelude Manager client.
@@ -134,15 +155,21 @@ static int send_auth_result(server_generic_client_t *client, int result)
 static int authenticate_client(server_generic_t *server, server_generic_client_t *client) 
 {
         int ret;
-        
-        if ( ! client->msg && ! (client->state & SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED) ) {
 
-                ret = manager_auth_client(client, client->fd);                
+        if ( ! client->msg && ! client->alert && ! (client->state & SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED) ) {
+                ret = manager_auth_client(client, client->fd, &client->alert);                
                 if ( ret == 0 )
                         return ret; /* EAGAIN happened */
                 
-                if ( ret < 0 )
-                        return send_auth_result(client, PRELUDE_MSG_AUTH_FAILED);
+                if ( ret < 0 ) {
+                        if ( client->alert ) {
+                                ret = send_queued_alert(client);
+                                if ( ret != 1 )
+                                        return ret;
+                        }
+                        
+                        return -1;
+                }
                 
                 client->state |= SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED;
                 
@@ -151,6 +178,12 @@ static int authenticate_client(server_generic_t *server, server_generic_client_t
                         return ret;
         }
 
+        else if ( client->alert ) {
+                ret = send_queued_alert(client);
+                if ( ret != 1 )
+                        return ret;
+        }
+        
         else if ( client->msg ) {
                 ret = send_auth_result(client, -1);
                 if ( ret != 1 )
@@ -205,15 +238,11 @@ static int read_connection_cb(void *sdata, server_logic_client_t *ptr)
         int ret = 0;
         server_generic_t *server = sdata;
         server_generic_client_t *client = (server_generic_client_t *) ptr;
-        
-        if ( client->state & SERVER_GENERIC_CLIENT_STATE_CLOSING ) {
-                /* stop further processing */
-                ret = (server_logic_remove_client(ptr) == 0) ? -2 : 0;
-        }
-        
-        else if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED )
+
+        assert(!(client->state & SERVER_GENERIC_CLIENT_STATE_CLOSING));
+                
+        if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED )
                 ret = server->read(client);
-        
         else
                 ret = authenticate_client(server, client);
         
@@ -222,6 +251,34 @@ static int read_connection_cb(void *sdata, server_logic_client_t *ptr)
 
 
 
+static int do_close_fd(server_generic_client_t *client)
+{
+        int ret;
+        void *fd_ptr;
+        prelude_error_code_t code;
+        
+        do {
+                ret = prelude_io_close(client->fd);                
+                if ( ret == 0 )
+                        break;
+
+                code = prelude_error_get_code(ret);
+                if ( code == PRELUDE_ERROR_EAGAIN ) {
+                        fd_ptr = prelude_io_get_fdptr(client->fd);
+                        if ( fd_ptr && gnutls_record_get_direction(fd_ptr) == 1 )
+                                server_logic_notify_write_enable((server_logic_client_t *) client);
+                        
+                        return -1;
+                }
+                        
+                server_generic_log_client(client, PRELUDE_LOG_WARN, "connection closure error: %s.\n", prelude_strerror(ret));
+                
+        } while ( ret < 0 && ! prelude_io_is_error_fatal(client->fd, ret));
+
+        return 0;
+}
+
+
 /*
  * callback called by server-logic when a connection should be closed.
  * if the authentication process succeed for this connection, call
@@ -232,7 +289,7 @@ static int close_connection_cb(void *sdata, server_logic_client_t *ptr)
         int ret;
         server_generic_t *server = sdata;
         server_generic_client_t *client = (server_generic_client_t *) ptr;
-
+        
         client->state |= SERVER_GENERIC_CLIENT_STATE_CLOSING;
         
         if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED && ! (client->state & SERVER_GENERIC_CLIENT_STATE_CLOSED) ) {
@@ -249,21 +306,14 @@ static int close_connection_cb(void *sdata, server_logic_client_t *ptr)
          * that they can take control over the connection FD.
          */
         if ( client->fd ) {
-                
-                ret = prelude_io_close(client->fd);
-                if ( ret < 0 && prelude_error_get_code(ret) == PRELUDE_ERROR_EAGAIN ) {
-
-                        if ( server->sa->sa_family != AF_UNIX && gnutls_record_get_direction(prelude_io_get_fdptr(client->fd)) == 1 )
-                                server_logic_notify_write_enable(ptr);
-                        
+                ret = do_close_fd(client);
+                if ( ret < 0 )
                         return -1;
-                }
                 
                 prelude_io_destroy(client->fd);
+                server_generic_log_client(client, PRELUDE_LOG_INFO, "closing connection.\n");
         }
         
-        server_generic_log_client(client, PRELUDE_LOG_INFO, "closing connection.\n");
-
         free(client->permission_string);
         free(client->addr);        
         free(client);
@@ -328,18 +378,16 @@ static int setup_client_socket(server_generic_t *server,
 #endif
         /*
          * set client socket non blocking.
-         */
+         */        
         ret = fcntl(client, F_SETFL, O_NONBLOCK);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not set non blocking mode for client: %s.\n", strerror(errno));
-                return -1;
-        }
+        if ( ret < 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "could not set non blocking mode for client: %s");
         
         fcntl(client, F_SETFD, fcntl(client, F_GETFD) | FD_CLOEXEC);
         
         ret = prelude_io_new(&cdata->fd);
         if ( ret < 0 ) 
-                return -1;
+                return ret;
 
         prelude_io_set_sys_io(cdata->fd, client);
                
@@ -354,8 +402,8 @@ static int setup_client_socket(server_generic_t *server,
 
 static int accept_connection(server_generic_t *server, server_generic_client_t *cdata) 
 {
-        int sock;
         socklen_t addrlen;
+        int ret, sock, on = 1;
         
 #ifndef HAVE_IPV6
         struct sockaddr_in addr;
@@ -370,7 +418,11 @@ static int accept_connection(server_generic_t *server, server_generic_client_t *
                 prelude_log(PRELUDE_LOG_ERR, "accept error: %s.\n", strerror(errno));
                 return -1;
         }
-                
+        
+        ret = setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(int));
+        if ( ret < 0 )
+                prelude_log(PRELUDE_LOG_ERR, "could not set SO_KEEPALIVE socket option: %s.\n", strerror(errno));
+
         if ( server->sa->sa_family == AF_UNIX )
                 cdata->addr = strdup(((struct sockaddr_un *) server->sa)->sun_path);
         else {
@@ -494,17 +546,14 @@ static int generic_server(int sock, struct sockaddr *addr, size_t alen)
         int ret;
         
         ret = bind(sock, addr, alen);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not bind socket: %s.\n", strerror(errno));
-                return -1;
-        }
+        if ( ret < 0 )
+                return prelude_error_verbose(prelude_error_code_from_errno(errno),
+                                             "could not bind socket: %s", strerror(errno));
         
         ret = listen(sock, 10);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not listen on socket: %s.\n", strerror(errno));
-                return -1;
-        }
-        
+        if ( ret < 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "listen error: %s", strerror(errno));
+                
         return 0;
 }
 
@@ -532,30 +581,22 @@ static int is_unix_socket_already_used(int sock, struct sockaddr_un *sa, int add
          */
         ret = stat(sa->sun_path, &st);
         if ( ret < 0 )
-                return FALSE;
+                return 0;
 
-        if ( ! S_ISSOCK(st.st_mode) ) {
-                prelude_log(PRELUDE_LOG_WARN, "%s already exist and is not an UNIX socket: please check.\n",
-                            sa->sun_path);
-                return -1;
-        }
+        if ( ! S_ISSOCK(st.st_mode) )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "'%s' exist and is not an UNIX socket", sa->sun_path);
         
         ret = connect(sock, (struct sockaddr *) sa, addrlen);
-        if ( ret == 0 ) {
-                prelude_log(PRELUDE_LOG_WARN, "%s UNIX socket is already in use. Exiting.\n",
-                            sa->sun_path);
-                return TRUE;
-        }
+        if ( ret == 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "UNIX socket '%s' already in use", sa->sun_path);
         
         /*
          * The unix socket exist on the file system,
          * but no one use it... Delete it.
          */
         ret = unlink(sa->sun_path);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "couldn't delete UNIX socket.\n");
-                return -1;
-        }
+        if ( ret < 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "could not delete unused UNIX socket: %s", strerror(errno));
         
         return 0;
 }
@@ -571,21 +612,19 @@ static int unix_server_start(server_generic_t *server)
         struct sockaddr_un *sa = (struct sockaddr_un *) server->sa;
         
         server->sock = socket(AF_UNIX, SOCK_STREAM, 0);
-        if ( server->sock < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "error creating UNIX socket: %s.\n", strerror(errno));
-		return -1;
-	}
+        if ( server->sock < 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "error creating UNIX socket: %s", strerror(errno));
 
         ret = is_unix_socket_already_used(server->sock, sa, server->slen);
-        if ( ret == 1 || ret < 0  ) {
+        if ( ret < 0 ) {
                 close(server->sock);
-                return -1;
+                return ret;
         }
 
         ret = generic_server(server->sock, server->sa, server->slen);
         if ( ret < 0 ) {
                 close(server->sock);
-                return -1;
+                return ret;
         }
 
         /*
@@ -593,10 +632,8 @@ static int unix_server_start(server_generic_t *server)
          * representing our socket.
          */
         ret = chmod(sa->sun_path, S_IRWXU|S_IRWXG|S_IRWXO);
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not set permission on UNIX socket: %s.\n", strerror(errno));
-                return -1;
-        }
+        if ( ret < 0 ) 
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "could not set permission on UNIX socket: %s", strerror(errno));
         
         return 0;
 }
@@ -612,20 +649,12 @@ static int inet_server_start(server_generic_t *server, struct sockaddr *addr, so
         int ret, on = 1;
         
         server->sock = socket(server->sa->sa_family, SOCK_STREAM, IPPROTO_TCP);
-        if ( server->sock < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not create socket: %s.\n", strerror(errno));
-                return -1;
-        }
+        if ( server->sock < 0 )
+                return prelude_error_verbose(PRELUDE_ERROR_GENERIC, "error creating socket: %s", strerror(errno));
         
         ret = setsockopt(server->sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int));
         if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not set SO_REUSEADDR socket option: %s.\n", strerror(errno));
-                goto err;
-        }
-
-        ret = setsockopt(server->sock, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(int));
-        if ( ret < 0 ) {
-                prelude_log(PRELUDE_LOG_ERR, "could not set SO_KEEPALIVE socket option: %s.\n", strerror(errno));
+                ret = prelude_error_verbose(PRELUDE_ERROR_GENERIC, "error setting SO_REUSEADDR: %s", strerror(errno));
                 goto err;
         }
         
@@ -637,7 +666,7 @@ static int inet_server_start(server_generic_t *server, struct sockaddr *addr, so
 
  err:
         close(server->sock);
-        return -1;
+        return ret;
 }
 
 
@@ -676,13 +705,11 @@ static int do_getaddrinfo(struct addrinfo **ai, const char *addr, unsigned int p
         hints.ai_protocol = IPPROTO_TCP;
                 
         ret = getaddrinfo(addr, service, &hints, ai);
-        if ( ret != 0 ) {
-                prelude_log(PRELUDE_LOG_WARN, "could not resolve %s: %s.\n",
-                            addr, (ret == EAI_SYSTEM) ? strerror(errno) : gai_strerror(ret));
-                return -1;
-        }
-
-        return 0;
+        if ( ret != 0 )
+                ret = prelude_error_verbose(PRELUDE_ERROR_GENERIC, "could not resolve '%s': %s",
+                                            addr, (ret == EAI_SYSTEM) ? strerror(errno) : gai_strerror(ret));
+                
+        return ret;
 }
 
 
@@ -701,7 +728,7 @@ static int resolve_addr(server_generic_t *server, const char *addr, unsigned int
         else {
                 ret = do_getaddrinfo(&ai, addr, port);
                 if ( ret < 0 )
-                        return -1;
+                        return ret;
 
                 ai_family = ai->ai_family;
                 ai_addrlen = ai->ai_addrlen;
@@ -709,9 +736,8 @@ static int resolve_addr(server_generic_t *server, const char *addr, unsigned int
 
         server->sa = malloc(ai_addrlen);
         if ( ! server->sa ) {
-                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 freeaddrinfo(ai);
-                return -1;
+                return prelude_error_from_errno(errno);
         }
 
         server->slen = ai_addrlen;
@@ -730,6 +756,28 @@ static int resolve_addr(server_generic_t *server, const char *addr, unsigned int
 
 
 
+static int sg_bind_common(server_generic_t *server, unsigned int port)
+{
+        char out[128];
+        void *in_addr;
+        
+        fcntl(server->sock, F_SETFD, fcntl(server->sock, F_GETFD) | FD_CLOEXEC);
+        
+        if ( server->sa->sa_family == AF_UNIX )
+                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s).\n",
+                            ((struct sockaddr_un *) server->sa)->sun_path);
+        else {
+                in_addr = prelude_sockaddr_get_inaddr(server->sa);
+                assert(in_addr);
+                
+                inet_ntop(server->sa->sa_family, in_addr, out, sizeof(out));
+                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s port %u).\n", out, port);
+        }
+
+        return 0;
+}
+
+
 
 /*
  *
@@ -765,11 +813,32 @@ server_generic_t *server_generic_new(size_t clientlen, server_generic_accept_fun
 
 
 
+int server_generic_bind_numeric(server_generic_t *server, struct sockaddr *sa, socklen_t len, unsigned int port)
+{
+        int ret;
+
+        server->sa = malloc(len);
+        if ( ! server->sa )
+                return prelude_error_from_errno(errno);
+
+        server->slen = len;
+        memcpy(server->sa, sa, len);
+        
+        ret = inet_server_start(server, server->sa, server->slen);
+        if ( ret < 0 ) {
+                server_logic_stop(server->logic);
+                free(server->sa);
+                return ret;
+        }
+
+        return sg_bind_common(server, port);
+}
+
+
+
 int server_generic_bind(server_generic_t *server, const char *saddr, unsigned int port)
 {
         int ret;
-        char out[128];
-        void *in_addr;
         
         ret = resolve_addr(server, saddr, port);
         if ( ret < 0 )
@@ -783,25 +852,10 @@ int server_generic_bind(server_generic_t *server, const char *saddr, unsigned in
         if ( ret < 0 ) {
                 server_logic_stop(server->logic);
                 free(server->sa);
-                free(server);
-                return -1;
+                return ret;
         }
-                
-        fcntl(server->sock, F_SETFD, fcntl(server->sock, F_GETFD) | FD_CLOEXEC);
-        
-        if ( server->sa->sa_family == AF_UNIX )
-                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s).\n",
-                            ((struct sockaddr_un *) server->sa)->sun_path);
-        else {
-                in_addr = prelude_sockaddr_get_inaddr(server->sa);
-                assert(in_addr);
-                
-                inet_ntop(server->sa->sa_family, in_addr, out, sizeof(out));
-                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s port %u).\n", out, port);
-        }
-                
 
-        return 0;
+        return sg_bind_common(server, port);
 }
 
 
@@ -838,6 +892,11 @@ void server_generic_process_requests(server_generic_t *server, server_generic_cl
 }
 
 
+void server_generic_remove_client(server_generic_t *server, server_generic_client_t *client)
+{
+        server_logic_remove_client((server_logic_client_t *) client);
+}
+
 
 void server_generic_log_client(server_generic_client_t *cnx, prelude_log_t priority, const char *fmt, ...)
 {
